// WARNING:  this file is automatically included by JHashTable.h.
//           DO NOT MAKE THIS .CPP FILE ITS OWN COMPILATION UNIT.
//
//  Modified by Curtis Krauskopf (the Database Managers, Inc., www.decompile.com)
//  for wider distribution and to apply modern programming practices while trying
//  to maintain the spirit of the original library.


template <class KeyType, class ObjType>
Hashtable<KeyType,ObjType>::Hashtable(
  unsigned int(*func)(const KeyType&),
  const ObjType &null_item,
  unsigned int initialCapacity /* = 101 */, float loadFactor /* = 0.5f */ )
  throw (char *)
  : m_count(0), m_tableSize(initialCapacity), m_loadFactor(loadFactor),
    m_threshold((unsigned int)(initialCapacity * loadFactor)),
    m_table(NULL), UserHash(NULL), NULL_ITEM(null_item)
{
    if ( loadFactor <= 0.0 || loadFactor > 1.0 )
    {
#ifdef JAVALIB_USE_CERR
        cerr << "Bad Argument for loadFactor" << endl;
        exit(1);
#else
        throw "Bad Argument for loadFactor";
#endif
    }

    //initialize and set to null
    //==========================
    m_table = new HashtableEntry*[m_tableSize];
    memset( m_table, 0, sizeof(HashtableEntry*) * m_tableSize );

    setHashFunction(func);
}

template <class KeyType, class ObjType>
Hashtable<KeyType,ObjType>::~Hashtable( )
{
    clear();
    delete [] m_table;
}

template <class KeyType, class ObjType>
bool
Hashtable<KeyType,ObjType>::contains( const ObjType& value ) const
{
    HashtableEntry** tab = m_table;
    for( unsigned int i = 0; i < m_tableSize; i++ )
    {
        for( HashtableEntry* e = tab[i]; e != NULL; e = e->next )
        {
            if ( IsEqual(e->value, value) )
                return true;
        }
    }

    return false;
}

// cdk: remove const specifier because FindEntryByKey is non-const
template <class KeyType, class ObjType>
bool
Hashtable<KeyType,ObjType>::containsKey( const KeyType& key )
{
    // cdk:  changed == to !=   July 2, 2006
    return FindEntryByKey( key ) != NULL;
}


template <class KeyType, class ObjType>
const ObjType&
Hashtable<KeyType,ObjType>::get( const KeyType& key )
{
    HashtableEntry* entry = FindEntryByKey( key );
    
    if ( entry != NULL )
        return entry->value;
    
    return NULL_ITEM; //if not found
}

template <class KeyType, class ObjType>
void
Hashtable<KeyType,ObjType>::rehash()
{
    unsigned int oldTableSize = m_tableSize;
    HashtableEntry** oldTable = m_table;

    unsigned int newTableSize = nextPrime( 2 * oldTableSize );
    HashtableEntry** newTable = new HashtableEntry*[newTableSize];
    
    memset( newTable, 0, sizeof(HashtableEntry*) * newTableSize );
    

    m_threshold = (unsigned int)(newTableSize * m_loadFactor);
    m_table = newTable;

    //copy all the entries from the old to the new table
    for(unsigned int i = 0; i < oldTableSize; i++ )
    {
        for( HashtableEntry* old = oldTable[i]; old != NULL; )
        {
            HashtableEntry* e = old;
            old = old->next;

            unsigned int index = e->hash % newTableSize;
            e->next = newTable[index];
            newTable[index] = e;
        }
    }

    m_tableSize = newTableSize;

    delete [] oldTable;
}

template <class KeyType, class ObjType>
ObjType
Hashtable<KeyType,ObjType>::put( const KeyType& key, const ObjType& value )
{
    //Makes sure the key is not already in the hashtable
    //replace item if found
    HashtableEntry* e = FindEntryByKey( key );    
    if ( e != NULL )
    {
            ObjType old = e->value;
            e->value = value;
            return old;
    }
    

    //Rehash the table if the threshold is exceeded
    if ( m_count >= m_threshold )
    {
        rehash();
        return put( key, value );
    }

    //Creates the new entry here
    e = new HashtableEntry();
    e->hash = hash;
    e->key = key;
    e->value = value;
    
    HashtableEntry** row = GetHashRow( key );
    e->next = *row;
    *row = e;
    m_count++;

    return NULL_ITEM;
}





// cdk: added these helper functions for testing when keys are equal.
// Function base template
template <typename T>
bool const IsEqual(T const &first, T const &second)
{
    return (first == second) ? true : false;
}

// Specialize for type 'char *'
template <>
inline bool IsEqual(char * first, char * second)
{
    return (std::strcmp(first, second) == 0);
}



template <class KeyType, class ObjType>
ObjType
Hashtable<KeyType,ObjType>::remove( const KeyType& key )
{
    HashtableEntry** hashRow = GetHashRow( key );

    for( HashtableEntry* e = *hashRow, *prev = NULL; e != NULL; prev = e, e = e->next)
    {
        if ( e->hash == hash && IsEqual(e->key,key) )
        {
            if ( prev != NULL )
                prev->next = e->next;
            else
                // cdk:  was:   tab[index] = tab[index]->next;
                // changed to below because of compile-time error:  July 1, 2006
                *hashRow = (*hashRow)->next;

            ObjType theReturn = e->value;
            delete e;
            m_count--;    //decrement counter
            return theReturn;
        }
    }

    return NULL_ITEM;
}

template <class KeyType, class ObjType>
void
Hashtable<KeyType,ObjType>::clear()
{
    HashtableEntry** tab = m_table;
    
    for(unsigned int i = 0; i < m_tableSize; i++ )
    {
        HashtableEntry* e = tab[i];
        while( e != NULL )
        {
            HashtableEntry* tmp = e;
            e = e->next;
            delete tmp;
        }
        
        tab[i] = NULL;
    }

    //reset the counter
    m_count = 0;
}

template <class KeyType, class ObjType>
unsigned int
Hashtable<KeyType,ObjType>::nextPrime( unsigned int start )
{
    if( start % 2 == 0 )
        start++;

    unsigned int i;

    for( ; ; start += 2 )
    {
        for( i = 3; i * i <= start; i += 2 )
            if( start % i == 0 )
                break;

        if( i * i > start )
            return start;
    }
}


template <class KeyType, class ObjType>
typename Hashtable<KeyType,ObjType>::HashtableEntry*
Hashtable<KeyType,ObjType>::FindEntryByKey( const KeyType& key )
{
    Hashtable<KeyType,ObjType>::HashtableEntry** hashRow = GetHashRow( key );

    for( HashtableEntry* e = *hashRow; e != NULL; e = e->next )
    {
        // cdk:  added IsEqual:  July 2, 2006
        if ( e->hash == hash && IsEqual(e->key, key) )
            return e;
    }

    return NULL;
}

template <class KeyType, class ObjType>
typename Hashtable<KeyType,ObjType>::HashtableEntry**
Hashtable<KeyType,ObjType>::GetHashRow( const KeyType& key )
{
    Hashtable<KeyType,ObjType>::HashtableEntry** tab = m_table;
    hash = (*UserHash)(key);
    unsigned int index = hash % m_tableSize;

    return &tab[index];
}

