/*
	Version 1.3  8/11/2005
  This library was downloaded from: http://www.mike95.com
  Modified by Curtis Krauskopf (the Database Managers, Inc., www.decompile.com)
  for wider distribution and to apply modern programming practices while trying
  to maintain the spirit of the original library.

  This library is copyright.  It may freely be used for personal purposes
  if the restriction listed below is adhered to.
      Author: Michael Olivero
      Email:  mike95@mike95.com

  //===============================
  //Start of Restriction Definition
  //===============================
  Anyone can have full use of the library provided they keep this complete comment
  with the source.  Also I would like to ask if any changes are made to the
  code for efficiency reasons, please let me know so I may look into your change and
  likewise incorporate it into library.  If a suggestion makes it into the library,
  your credits will be added to this information.

  Authors of Computer related books are welcome to include this source code as part
  of their publishing, provided credit the Author and make note of where the source
  code was obtained from: http://www.mike95.com
  //=============================
  //End of Restriction Definition
  //=============================

  Description:
  Visit http://www.mike95.com/c_plusplus/classes/JHashtable/

  This class is a based on the Java JHashtable class and as such contains all the public
  member functions of it's Java equivalent.  Unlike Java, typecasts are not necessary
  since C++ allows template enstatiation of types at compile time.

  Note:  Since java has hashItem() as a member of the base Object class, all 
  Java classes are inheritly hashable.  Since the template parameter types do
  not necessarily have to have a built in hashing function, the user of the class
  must specify a hash function by calling setHashFunction() passing a pointer 
  to the hash function.

  The has function must be declared as the following:

      UINT function( const KeyType& );

  Where:
    function = any name you choose to use for the function name
    KeyType  = the type used for the key in the construction of the JHashtable object.

  Example:
      UINT myHash( const int& )
      {
          //your hashing code here for a key of type int.
      }


  //The following people have contributed to the solution
  //of bugs or additional features in this library
  //=====================================================
  //Jeremy Friesner, email: jaf@chem.ucsd.edu
  //Roxana Arama, email:roxanaa@tlc.ro
  //Hongbin Huang <huang_hongbin@yahoo.com>
  //cdk (Curtis Krauskopf), email: curtis@decompile.com
*/
#ifndef COM_MIKE95_JHASHTABLE_H

#include <string.h>   // for strcmp

namespace Java
{
  template <class KeyType, class ObjType>
  class Hashtable
  {
  public:
    Hashtable( unsigned int(*func)(const KeyType&), const ObjType &null_item,
          unsigned int initialCapacity = 101, float loadFactor = 0.5f )
      throw (char *);
    virtual ~Hashtable();

    //Inspectors
    //==========
    unsigned int size() const { return m_count; }
    bool isEmpty() const { return m_count == 0; }
    bool contains( const ObjType& value ) const;
    bool containsKey( const KeyType& key );
    const ObjType& get( const KeyType& key );


    //Modifiers
    //=========
    void rehash();
    ObjType put( const KeyType& key, const ObjType& value );
    ObjType remove( const KeyType& key );
    void clear();

    //C++ specific user hash function
    void setHashFunction( unsigned int(*func)(const KeyType& key) ) { UserHash = func; }
    void setNullItem(const ObjType& value) { NULL_ITEM = value; }

    ObjType NULL_ITEM;    //used for returns of not found

  private:
    //type for Entries
    struct HashtableEntry
    {
        unsigned int hash;
        KeyType key;
        ObjType value;
        HashtableEntry* next;
    };

    //Member variables
    unsigned int m_count;    //the size of the elements in the hashtable
    unsigned int m_tableSize;    //the size of the table.
    unsigned int m_threshold;
    float m_loadFactor;
    HashtableEntry** m_table;


    //Helper functions
    HashtableEntry* FindEntryByKey( const KeyType& key );
    HashtableEntry** GetHashRow( const KeyType& key );


    unsigned int nextPrime( unsigned int start );
    unsigned int (*UserHash)( const KeyType& key );

    unsigned int hash;            // most recent hash code for most recent key
  };

  #include "JHashTable.cpp"
}  // namespace Java

#endif

