/*
   This library was downloaded from: http://www.mike95.com

   Modified by Curtis Krauskopf (the Database Managers, Inc., www.decompile.com)
   for wider distribution and to apply modern programming practices while trying
   to maintain the spirit of the original library.

   This library is copyright.  It may freely be used for personal purposes
   if the restriction listed below is adhered to.
       Author: Michael Olivero
       Email:  mike95@mike95.com

   //===============================
   //Start of Restriction Definition
   //===============================
   Anyone can have full use of the library provided they keep this complete comment
   with the source.  Also I would like to ask if any changes are made to the
   code for efficiency reasons, please let me know so I may look into your change and
   likewise incorporate it into library.  If a suggestion makes it into the library,
   your credits will be added to this information.

   Authors of Computer related books are welcome to include this source code as part
   of their publishing, provided credit the Author and make note of where the source
   code was obtained from: http://www.mike95.com
   //=============================
   //End of Restriction Definition
   //=============================


   Description:
   Visit http://www.mike95.com/c_plusplus/classes/JString/

   This library avoids the need to use pointers to char or manual
   memory manipulation while working with JStrings.  This class has been
   based on the Java String class, and thus has all of the Java public functionality
   with a few extras [replace()] needed for useful functionality.

   //The following people have contributed to the solution
   //of bugs or improvements in this library
   //=====================================================
   //Carl Pupa, [pumacat@erols.com]
   //Thomas Watson, [w@tson.dk]
   //Subbiah, Venkat [vsubbiah@corvis.com]
   //Oren Tirosh [mailto:oren@hishome.net]
   //Curtis Krauskopf [mailto: curtis@decompile.com]
*/

#ifndef __JString
#define __JString

#include <iostream.h>          // for ostream, istream

namespace Java {
  class String {
  public:
	// Constructor(s) / Destructors
	//-----------------------------
	String( const char *Value = "" );
	explicit String( const String &Value );
        explicit String( const char );
        explicit String( const unsigned char );
        explicit String( const short );
        explicit String( const unsigned short );
        explicit String( const int);
        explicit String( const unsigned int);
        explicit String( const long );
        explicit String( const unsigned long );
	virtual ~String() { if( BufferLen > 0 ) delete [] Buffer; }

	//Operators
	//----------
	const String &	operator = ( const String &Rhs );
	const String &	operator +=( const String &Rhs );
	const String &	operator +=( const char );
	int		operator ==( const String &Rhs ) const;
	int		operator !=( const String &Rhs ) const;
	int		operator < ( const String &Rhs ) const;
	int		operator > ( const String &Rhs ) const;
	int		operator <=( const String &Rhs ) const;
	int		operator >=( const String &Rhs ) const;
	char		operator []( unsigned int Index ) const;
	char&		operator []( unsigned int Index );

	//Methods
	//-------
	//INSPECTORS
	//==========
	char		charAt( unsigned int index ) const;
	int		compareTo( const String &anotherString ) const;
	const char*	cstr( ) const { return Buffer; }
	int		endsWith( const String &suffix ) const;
	int		equals( const String &anObject ) const;
	int		equalsIgnoreCase( const String &anotherString ) const;
	int		indexOf( char ch ) const;
	int		indexOf( char ch, unsigned int fromIndex ) const;
	int		indexOf( const String &str ) const;
	int		indexOf( const String &str, unsigned int fromIndex ) const;
	int		lastIndexOf( char ch ) const;
	int		lastIndexOf( char ch, unsigned int fromIndex ) const;
	int		lastIndexOf( const String &str ) const;
	int		lastIndexOf( const String &str, unsigned int fromIndex ) const;
	const unsigned int length( ) const { return Length; }
	int		startsWith( const String &prefix ) const;
	int		startsWith( const String &prefix, unsigned int toffset ) const;
	String		substring( unsigned int beginIndex ) const;
	String		substring( unsigned int beginIndex, unsigned int endIndex ) const;
	String		toLowerCase( ) const;
	String		toUpperCase( ) const;
	String		trim( ) const;

        // CDK added, 12/04/2002:
        const char *c_str() const { return Buffer; }

	//Methods
	//-------
	//MODIFIERS
	//=========
	const String&	concat( const String &str );
	String			replace( char oldChar, char newChar );
	String			replace( const String& match, const String& replace );
	// void			Format( const char* fmt, ...);

	// Friends
	//--------
	friend Java::String 	operator + ( Java::String Lhs, const Java::String &Rhs );
	friend ostream&		operator<< ( ostream &Out, const Java::String &Value );
	friend istream&		operator>> ( istream &In, Java::String &Value );

  protected:
	//Members
	//-------
	char *Buffer;			   // Stores the chars
	unsigned int BufferLen;			  // Max strlen for Buffer
	unsigned int Length;				 // Length of string

	void GetBuffer(unsigned int MaxStrLen);
	void Double( );
	void verifyIndex( unsigned int number ) const;
  };
}     // namespace Java::


//Class Functions
//===============
inline void
Java::String::GetBuffer(unsigned int MaxStrLen)
{
	BufferLen = MaxStrLen;
	Buffer = new char[BufferLen + 1];
}

inline void
Java::String::Double( )
{
	char *temp = Buffer;
	GetBuffer( ++BufferLen * 2 );
	strcpy( Buffer, temp );
	delete [] temp;
}

inline void
Java::String::verifyIndex( unsigned int index ) const
{
        // CDK commented out part of the following test because it's always false
	if ( /* index < 0 ||*/ index >= Length )
	{
#               ifdef JAVALIB_USE_CERR
		cerr << "Index Out Of Bounds Exception at ["
			<< index << "] in:\n" << Buffer << endl;
		exit(1);
#               else
		throw "Index Out Of Bounds Exception";
#               endif
	}
}

//Friend Functions
//================
inline Java::String
Java::operator+( Java::String Lhs, const Java::String &Rhs )
{
	return Lhs += Rhs;
}

#endif
