/*
  This library was downloaded from: http://www.mike95.com

  Modified by Curtis Krauskopf (the Database Managers, Inc., www.decompile.com)
  for wider distribution and to apply modern programming practices while trying
  to maintain the spirit of the original library.

  This library is copyright.  It may freely be used for personal purposes
  if the restriction listed below is adhered to.
      Author: Michael Olivero
      Email:  mike95@mike95.com

  //===============================
  //Start of Restriction Definition
  //===============================
  Anyone can have full use of the library provided they keep this complete comment
  with the source.  Also I would like to ask if any changes are made to the
  code for efficiency reasons, please let me know so I may look into your change and
  likewise incorporate it into library.  If a suggestion makes it into the library,
  your credits will be added to this information.

  Authors of Computer related books are welcome to include this source code as part
  of their publishing, provided credit the Author and make note of where the source
  code was obtained from: http://www.mike95.com
  //=============================
  //End of Restriction Definition
  //=============================

  Description:
  Visit http://www.mike95.com/c_plusplus/classes/JVector/

  Standard collection class.  It's public member functions
  are identical to the Java Vector public member functions (except for any Java specific
  Java related functions).

  //The following people have contributed to the solution
  //of bugs or additional features in this library
  //=====================================================
  //Carl Pupa, email: pumacat@erols.com
  //Adam Doppelt, email: amd@gurge.com
  //Dave O'Keefe, email: okeefe@cs.stanford.edu
  //George Kerber, email: gek@pobox.com
  //Curtis Krauskopf, email: curtis@decompile.com
*/

// CDK:  added throw() specifiers:  July 1, 2006

#ifndef COM_MIKE95_COM_JVECTOR_H
#define COM_MIKE95_COM_JVECTOR_H

// #include <iostream>

namespace Java {
  template <class Etype>
  class Vector
  {
  public:
    Vector( unsigned int initialCapacity = 10, unsigned int capacityIncrement = 10 );
    Vector( const Vector& rhv );
    virtual ~Vector();

    //Inspectors (additional exception throwing inspectors below)
    //===========================================================
    unsigned int capacity() const;
    bool contains( const Etype& elem ) const;
    const Etype& firstElement() const
      throw(char *);
    int indexOf( const Etype& elem ) const;
    bool isEmpty() const;
    const Etype& lastElement() const
      throw(char *);
    int lastIndexOf( const Etype& elem ) const;
    unsigned int size() const;
    void copyInto( Etype* array ) const;

    //Modifiers (additional exception throwing inspectors below)
    //==========================================================
    inline bool add( const Etype& obj ) { addElement(obj);  return true; }
    void addElement( const Etype& obj );
    inline void clear() { removeAllElements(); }
    void ensureCapacity( unsigned int minCapacity );
    void removeAllElements();
    bool removeElement( const Etype& obj );

    void setSize( unsigned int newSize );
    void trimToSize();

    //Exceptions are thrown at run time with the following functions if
    //the index parameter is not within a valid range.  If the data
    //is uncertain (i.e. user inputed), then you should wrap these function
    //calls with the try/catch blocks and handle them appropriately.
    //===============================================
    const Etype& elementAt( unsigned int index ) const
      throw (char *);          //inspector
    void insertElementAt( const Etype& obj, unsigned int index )
      throw (char *);          //modifier
    const Etype& remove( unsigned int index )
      throw (char *);          //modifier
    void removeElementAt( unsigned int index )
      throw (char *);          //modifier
    void setElementAt( const Etype& obj, unsigned int index )
      throw (char *);          //modifier
    inline const Etype& get( unsigned int index ) const
      throw (char *)           // inspector
      { return elementAt(index); }


    //C++ specific operations
    //=======================
    const Etype& operator[]( unsigned int index ) const;
    Etype& operator[]( unsigned int index )
      throw (char *);

  protected:
    int min( unsigned int left, unsigned int right ) const;
    void verifyIndex( unsigned int index ) const
      throw (char *);
    unsigned int m_size;
    unsigned int m_capacity;
    unsigned int m_increment;
    Etype** m_pData;
  };
}    // namespace Java

#include "JVector.cpp"   // for template

#endif

