/*******************************************
   Downloaded from: http://www.mike95.com
   Copyright (c)1997 Michael Olivero
   All Rights Reserved

  Modified by Curtis Krauskopf (the Database Managers, Inc., www.decompile.com)
  for wider distribution and to apply modern programming practices while trying
  to maintain the spirit of the original library.

********************************************/
#include "JString.h"


//============
//Constructors
//============
Java::String::String( const char *Value /* = "" */ )
{
	if ( Value == NULL )
		Value = "";

	GetBuffer( Length = strlen( Value ) );
	strcpy( Buffer, Value );
}

Java::String::String( const Java::String &Value )
{
	GetBuffer( Length = Value.Length );
	strcpy( Buffer, Value.Buffer );
}



// Convert the char into a string
Java::String::String( const char value)
{
  Length = 1;
  GetBuffer(1);
  Buffer[0] = value;
  Buffer[1] = 0;
}


// Convert the char into a string
Java::String::String( const unsigned char value)
{
  Length = 1;
  GetBuffer(1);
  Buffer[0] = value;
  Buffer[1] = 0;
}


// Convert the int into a string
Java::String::String( const short value)
{
  char buf[33];   // longest possible return from itoa
  itoa((int)value, buf, 10);
  GetBuffer( Length = strlen(buf) );
  strcpy( Buffer, buf );
}


// Convert the int into a string
Java::String::String( const unsigned short value)
{
  char buf[33];   // longest possible return from ultoa
  ultoa((unsigned long)value, buf, 10);
  GetBuffer( Length = strlen(buf) );
  strcpy( Buffer, buf );
}

// Convert the long into a string
Java::String::String( const int value )
{
  char buf[33];   // longest possible return from itoa
  itoa((signed long)value, buf, 10);
  GetBuffer( Length = strlen(buf) );
  strcpy( Buffer, buf );
}

// Convert the long into a string
Java::String::String( const unsigned int value )
{
  char buf[33];   // longest possible return from ultoa
  ultoa((unsigned long)value, buf, 10);
  GetBuffer( Length = strlen(buf) );
  strcpy( Buffer, buf );
}


// Convert the long into a string
Java::String::String( const long value )
{
  char buf[33];   // longest possible return from ltoa
  ltoa(value, buf, 10);
  GetBuffer( Length = strlen(buf) );
  strcpy( Buffer, buf );
}

// Convert the long into a string
Java::String::String( const unsigned long value )
{
  char buf[33];   // longest possible return from ultoa
  ultoa(value, buf, 10);
  GetBuffer( Length = strlen(buf) );
  strcpy( Buffer, buf );
}



char
Java::String::charAt( unsigned int loc ) const
{
	return operator[]( loc );
}

int
Java::String::compareTo( const String &s2 ) const
{
	return strcmp( Buffer, s2.Buffer );
}

const Java::String &
Java::String::concat( const String &s2 )
{
	return (*this) += s2;
}

const Java::String &
Java::String::operator=( const String &Rhs )
{
	if ( this == &Rhs )
		return *this;

	if ( Rhs.Length > Length )
	{
		delete [] Buffer;
		GetBuffer( Rhs.Length );
	}

	Length = Rhs.Length;
	strcpy( Buffer, Rhs.Buffer );

	return *this;
}

const Java::String &
Java::String::operator+=( const char aChar )
{
	if ( Length == BufferLen )
		Double();

	Buffer[ Length++ ] = aChar;
	Buffer[ Length ] = '\0';

	return *this;
}

const Java::String &
Java::String::operator+=( const Java::String &other )
{
	Length += other.Length;
	if ( Length > BufferLen )
	{
		char *temp = Buffer;
		GetBuffer( Length );
		strcpy( Buffer, temp );
		delete [] temp;
	}
	strcat( Buffer, other.Buffer );

	return *this;
}


int
Java::String::operator==( const Java::String &Rhs ) const
{
	return ( Length == Rhs.Length && strcmp( Buffer, Rhs.Buffer ) == 0 );
}

int
Java::String::operator!=( const Java::String &Rhs ) const
{
	return ( Length != Rhs.length() || strcmp( Buffer, Rhs.cstr() ) != 0 );
}

int
Java::String::operator<( const Java::String &Rhs ) const
{
	return strcmp( Buffer, Rhs.Buffer ) < 0;
}

int
Java::String::operator>( const Java::String &Rhs ) const
{
	return strcmp( Buffer, Rhs.Buffer ) > 0;
}

int
Java::String::operator<=( const Java::String &Rhs ) const
{
	return strcmp( Buffer, Rhs.Buffer ) <= 0;
}

int
Java::String::operator>=( const Java::String & Rhs ) const
{
	return strcmp( Buffer, Rhs.Buffer ) >= 0;
}

char &
Java::String::operator[]( unsigned int Index )
{
	verifyIndex( Index );
	return Buffer[ Index ];
}

char
Java::String::operator[]( unsigned int Index ) const
{
	verifyIndex( Index );
	return Buffer[ Index ];
}


int
Java::String::endsWith( const Java::String &s2 ) const
{
	if ( Length < s2.Length )
		return 0;

	return strcmp( &Buffer[ Length - s2.Length], s2.cstr() ) == 0;
}

int
Java::String::equals( const Java::String &s2 ) const
{
	return ( Length == s2.Length && strcmp( Buffer,s2.Buffer ) == 0 );
}


int
Java::String::equalsIgnoreCase( const Java::String &s2 ) const
{
	if ( this == &s2 )
		return 1;
	else if ( Length != s2.Length )
		return 0;

	return strcmp(toLowerCase().cstr(), s2.toLowerCase().cstr()) == 0;
}


/* CDK Commented out this function because it's doing nothing
       useful and because the compiler was complaining that iterator was
       assigned but not used.  -- 12/06/2002
void
Java::String::Format( const char*, ... )
{
	va_list iterator;
	va_start( iterator, fmt );


	va_end( iterator );
}
*/


Java::String
Java::String::replace( char findChar, char replaceChar )
{
	Java::String theReturn = Buffer;
	char* temp = theReturn.Buffer;
        // CDK:  rewritten to suppress a compile-time warning 12/06/2002
	while( (temp = strchr( temp, findChar )) != 0 )
		*temp = replaceChar;

	return theReturn;
}

Java::String
Java::String::replace( const Java::String& match, const Java::String& replace )
{
	Java::String temp = Buffer, newString;

	int loc;
	while ( (loc = temp.indexOf( match )) != -1 )
	{
		newString += temp.substring( 0, loc );
		newString += replace;
		temp = temp.substring( loc + match.Length );
	}

	newString += temp;				   //get rest of string;

	return newString;
}

int
Java::String::indexOf( char temp ) const
{
	return indexOf( temp, 0 );
}


// Return the index of the specified character, or -1 if
// the character is not in the string.
// Even the null character can be searched.
//
int
Java::String::indexOf( char ch, unsigned int fromIndex ) const
{
	if ( fromIndex >= Length )
		return -1;

	const char* temp = strchr( &Buffer[fromIndex], ch );
	if ( temp == NULL )
		return -1;

	return temp - Buffer;
}

int
Java::String::indexOf( const Java::String &s2 ) const
{
	return indexOf( s2, 0 );
}


int
Java::String::indexOf( const Java::String &s2, unsigned int fromIndex ) const
{
	if ( fromIndex >= Length )
		return -1;

	const char *theFind = strstr( &Buffer[ fromIndex ], s2.cstr() );

	if ( theFind == NULL )
		return -1;

	return theFind - Buffer; //pointer subtraction
}



int
Java::String::lastIndexOf( char theChar ) const
{
	return lastIndexOf( theChar, Length - 1 );
}

int
Java::String::lastIndexOf( char ch, unsigned int fromIndex ) const
{
	if ( fromIndex >= Length )
		return -1;

	char tempchar = Buffer[fromIndex + 1];
	Buffer[fromIndex + 1] = '\0';
	char* temp = strrchr( Buffer, ch );
	Buffer[fromIndex + 1] = tempchar;

	if ( temp == NULL )
		return -1;

	return temp - Buffer;
}

int
Java::String::lastIndexOf( const Java::String &s2 ) const
{
	return lastIndexOf( s2, Length - s2.Length );
}

int
Java::String::lastIndexOf( const Java::String &s2, unsigned int fromIndex ) const
{
	//=============================
	//avoid check for empty strings
	//=============================
	if ( s2.Length == 0 || s2.Length - 1 > fromIndex || 
		fromIndex >= Length )
	return -1;

	//========================
	//matching first character
	//========================
	char temp = s2[ 0 ];

	for ( int i = fromIndex; i >= 0; i-- )
	{
		if ( Buffer[ i ] == temp &&
			(*this).substring( i, i + s2.Length ).equals( s2 ) )
			return i;
	}
	return -1;
}

int
Java::String::startsWith( const Java::String &s2 ) const
{
	if ( Length < s2.Length )
		return 0;

	return startsWith( s2, 0 );
}

int
Java::String::startsWith( const Java::String &s2, unsigned int offset ) const
{
	if ( offset > Length - s2.Length )
		return 0;

	return strncmp( &Buffer[offset], s2.cstr(), s2.Length ) == 0;
}


// Create a substring starting at the 0-based index position
Java::String
Java::String::substring( unsigned int left ) const
{
	return substring( left, Length );
}


// Create a substring starting at left and continuing to the character
// before right.  Both left and right are 0-based index pointers.
// Right can point to the terminating null character of the string.
// Note:  if left points 'to the right' of right, then left and
// right are swapped so that the left pointer is always to the
// left of right.
// If left == right, substring returns an empty string.
//
Java::String
Java::String::substring( unsigned int left, unsigned int right ) const
{
	if ( left > right )
	{
		int temp = right;
		right = left;
		left = temp;
	}

	if ( right > Length )
	{
                // CDK cerr is now compile-time dependent so that this
                // library does not always depend on cerr.
#ifdef JAVALIB_USE_CERR
                cerr << "Index Out Of Bounds Exception w/ substring(" << left
			<< "," << right << "]:\n" << Buffer << endl;
		exit(1);
#else
                // CDK added new behavior to adjust 'right' at runtime:
                right = Length;
#endif
	}

	char temp = Buffer[ right ]; //save the replaced character
	Buffer[ right ] = '\0';	  //nullify the the character

	Java::String outPut = ( Buffer + left ); //Pointer arithmetic

	Buffer[ right ] = temp;	  //restore character

	return outPut;
}

Java::String
Java::String::toLowerCase( ) const
{
	Java::String temp = Buffer;

	for ( unsigned int i = 0; i < Length; i++ )
                // CDK added typecast to suppress compiler warning, 12/06/2002
		temp.Buffer[ i ] = (char)tolower( temp.Buffer[ i ] );

	return temp;
}

Java::String
Java::String::toUpperCase() const
{
	Java::String temp = Buffer;

	for ( unsigned int i = 0; i < Length; i++ )
                // CDK added typecast to suppress compiler warning, 12/06/2002
		temp.Buffer[ i ] = (char)toupper( temp.Buffer[ i ] );

	return temp;
}

Java::String
Java::String::trim() const
{
	Java::String temp = Buffer;
	unsigned int i,j;

	for ( i = 0; i < Length; i++ )
	{
		if ( !isspace(Buffer[i]) )
			break;
	}

	for ( j = temp.Length - 1; j > i; j-- )
	{
		if ( !isspace(Buffer[j]) )
			break;
	}

	return temp.substring( i, j + 1);
}

istream &
Java::operator >> ( istream &In, Java::String &Value )
{
	static char Str[ 2048 ]; // allocate max size of 2048 characters
	In >> Str;
	Value = Str;			 // assign to reference Java::String
	return In;			   // Return istream
}


ostream&
Java::operator<< ( ostream &Out, const Java::String &Value )
{
        Out << Value.Buffer;
        return Out;
}

